from cython cimport view
import numpy as np
cimport numpy as np

cdef extern from "bitstream.h":
  cdef struct bitstream:
    pass

  cdef bitstream* stream_open(void* buffer, size_t bytes);

  cdef void stream_close(bitstream* stream);

cdef extern from "zfp/types.h":
  ctypedef unsigned int uint

cdef extern from "zfp.h":
  cdef double zfp_stream_set_accuracy(zfp_stream* stream, double tolerance);
  ctypedef enum zfp_exec_policy:
    zfp_exec_serial = 0, #/* serial execution (default) */
    zfp_exec_omp    = 1, #/* OpenMP multi-threaded execution */
    zfp_exec_cuda   = 2  #/* CUDA parallel execution */

  ctypedef enum zfp_type:
    zfp_type_none   = 0, #/* unspecified type */
    zfp_type_int32  = 1, #/* 32-bit signed integer */
    zfp_type_int64  = 2, #/* 64-bit signed integer */
    zfp_type_float  = 3, #/* single precision floating point */
    zfp_type_double = 4  #/* double precision floating point */

  #/* OpenMP execution parameters */
  ctypedef struct zfp_exec_params_omp:
    uint threads;    #/* number of requested threads */
    uint chunk_size; #/* number of blocks per chunk (1D only) */

  #/* execution parameters */
  ctypedef union zfp_exec_params:
    zfp_exec_params_omp omp; #/* OpenMP parameters */

  ctypedef struct zfp_execution:
    zfp_exec_policy policy; #/* execution policy (serial, omp, ...) */
    zfp_exec_params params; #/* execution parameters */

  ctypedef struct zfp_stream:
    pass
    #uint minbits;      # /* minimum number of bits to store per block */
    #uint maxbits;      # /* maximum number of bits to store per block */
    #uint maxprec;      # /* maximum number of bit planes to store */
    #int minexp;        # /* minimum floating point bit plane number to store */
    #bitstream* stream; # /* compressed bit stream */
    #zfp_execution exec1;# /* execution policy and parameters */

  ctypedef struct zfp_field:
      zfp_type type;       #/* scalar type (e.g. int32, double) */
      uint nx, ny, nz, nw; #/* sizes (zero for unused dimensions) */
      int sx, sy, sz, sw;  #/* strides (zero for contiguous array a[nw][nz][ny][nx]) */
      void* data;          #/* pointer to array data */

  cdef uint zfp_stream_set_precision(zfp_stream* stream, uint precision);

  cdef double zfp_stream_set_rate(
      zfp_stream* stream, #/* compressed stream */
      double rate,        #/* desired rate in compressed bits/scalar */
      zfp_type stype,      #/* scalar type to compress */
      uint dims,          #/* array dimensionality (1, 2, or 3) */
      int wra             #/* nonzero if write random access is needed */
  );

  cdef size_t zfp_stream_maximum_size(
      const zfp_stream* stream, #/* compressed stream */
      const zfp_field* field    #/* array to compress */
  );

  cdef void zfp_stream_set_bit_stream(
      zfp_stream* stream, #/* compressed stream */
      bitstream* bs       #/* bit stream to read from and write to */
  );

  cdef void zfp_stream_rewind(
      zfp_stream* stream #/* compressed bit stream */
  );

  cdef zfp_stream* zfp_stream_open(
      bitstream* stream #/* bit stream to read from and write to (may be NULL) */
  );

  cdef void zfp_stream_close(zfp_stream* stream);

  cdef int zfp_stream_set_execution(
      zfp_stream* stream,    #/* compressed stream */
      zfp_exec_policy policy #/* execution policy */
  );

  cdef zfp_field* zfp_field_1d(
    void* pointer, #/* pointer to uncompressed scalars (may be NULL) */
    zfp_type type, #/* scalar type */
    uint nx        #/* number of scalars */
   );

  cdef zfp_field* zfp_field_2d(
    void* pointer, #/* pointer to uncompressed scalars (may be NULL) */
    zfp_type type, #/* scalar type */
    uint nx,       #/* number of scalars in x dimension */
    uint ny        #/* number of scalars in y dimension */
   );

  cdef zfp_field* zfp_field_3d(
    void* pointer, #/* pointer to uncompressed scalars (may be NULL) */
    zfp_type type, #/* scalar type */
    uint nx,       #/* number of scalars in x dimension */
    uint ny,       #/* number of scalars in y dimension */
    uint nz        #/* number of scalars in z dimension */
   );

  cdef zfp_field* zfp_field_4d(
    void* pointer, #/* pointer to uncompressed scalars (may be NULL) */
    zfp_type type, #/* scalar type */
    uint nx,       #/* number of scalars in x dimension */
    uint ny,       #/* number of scalars in y dimension */
    uint nz,       #/* number of scalars in z dimension */
    uint nw        #/* number of scalars in w dimension */
   );

  cdef size_t zfp_compress(
      zfp_stream* stream,    #/* compressed stream */
      const zfp_field* field #/* field metadata */
  );

  cdef size_t zfp_decompress(
      zfp_stream* stream, #/* compressed stream */
      zfp_field* field    #/* field metadata */
  );

  cdef void zfp_field_free(
      zfp_field* field #/* field metadata */
  );

cdef void* raw_pointer_double(arr) except NULL:
    assert(arr.dtype==np.float64)
    assert(arr.flags.c_contiguous) # if this isn't true, ravel will make a copy
    cdef double[::1] mview = arr.ravel()
    return <void*>&mview[0]

cdef void* raw_pointer_float(arr) except NULL:
    assert(arr.flags.c_contiguous) # if this isn't true, ravel will make a copy
    assert(arr.dtype == np.float32)
    cdef float[::1] mview = arr.ravel()
    return <void*>&mview[0]

cdef void* raw_pointer(arr):
    if arr.dtype == np.float32:
        return raw_pointer_float(arr)
    else:
        return raw_pointer_double(arr)

zfp_types = {np.dtype('float32'): zfp_type_float, np.dtype('float64'): zfp_type_double}

cdef zfp_field* init_field(np.ndarray indata):
    data_type = zfp_types[indata.dtype]
    numdims = len((<object> indata).shape)
    shape = indata.shape
    if numdims == 1:
        return zfp_field_1d(raw_pointer(indata), data_type, shape[0])
    elif numdims == 2:
        return zfp_field_2d(raw_pointer(indata), data_type, shape[1], shape[0])
    elif numdims == 3:
        return zfp_field_3d(raw_pointer(indata), data_type, shape[2], shape[1], shape[0])
    #elif numdims == 4:
    #    return zfp_field_4d(raw_pointer(indata), data_type, shape[0], shape[1], shape[2], shape[4])
    else:
        raise ValueError("Invalid number of dimensions (valid: 1-4)")

def compress(indata, tolerance=None, precision=None, rate=None, parallel=True):
    assert(tolerance or precision or rate)
    assert(not(tolerance is not None and precision is not None))
    assert(not(tolerance is not None and rate is not None))
    assert(not(rate is not None and precision is not None))

    shape = list(reversed(indata.shape))

    field = init_field(indata)
    stream = zfp_stream_open(NULL)
    data_type = zfp_types[indata.dtype]
    # TODO Use return value to ensure we succeeded in all these 'set' calls
    if tolerance is not None:
        zfp_stream_set_accuracy(stream, tolerance)
    elif precision is not None:
        zfp_stream_set_precision(stream, precision)
    elif rate is not None:
        zfp_stream_set_rate(stream, rate, data_type, len(shape), 0)
    # Try multithreaded
    if(parallel):
        zfp_stream_set_execution(stream, zfp_exec_omp)
    bufsize = zfp_stream_maximum_size(stream, field)
    cdef char[::1] buff = view.array(shape=(bufsize,), itemsize=sizeof(char), format='B')
    bitstream = stream_open(<void *>&buff[0], bufsize)
    zfp_stream_set_bit_stream(stream, bitstream)
    zfp_stream_rewind(stream)
    zfpsize = zfp_compress(stream, field)

    zfp_field_free(field)
    zfp_stream_close(stream)
    stream_close(bitstream)
    return buff[:zfpsize]


def decompress(const unsigned char[::1] compressed, shape, dtype, tolerance=None, precision=None,
               rate=None, parallel=True):
    assert(tolerance or precision or rate)
    assert(not(tolerance is not None and precision is not None))
    assert(not(tolerance is not None and rate is not None))
    assert(not(rate is not None and precision is not None))
    outdata = np.zeros(shape, dtype=dtype)
    data_type = zfp_types[dtype]
    shape = list(reversed(shape))
    field = init_field(outdata)
    stream = zfp_stream_open(NULL)

    if tolerance is not None:
        zfp_stream_set_accuracy(stream, tolerance)
    elif precision is not None:
        zfp_stream_set_precision(stream, precision)
    elif rate is not None:
        zfp_stream_set_rate(stream, rate, data_type, len(shape), 0)
    # Try multithreaded
    #if(parallel):
    #    zfp_stream_set_execution(stream, zfp_exec_omp)
    bitstream = stream_open(<void*>&compressed[0], len(compressed))
    zfp_stream_set_bit_stream(stream, bitstream)
    zfp_stream_rewind(stream)
    zfp_decompress(stream, field)
    zfp_field_free(field)
    zfp_stream_close(stream)
    stream_close(bitstream)
    return outdata



# Thread and interpreter state structures and their interfaces

from .object cimport PyObject

cdef extern from "Python.h":

    # We make these an opaque types. If the user wants specific attributes,
    # they can be declared manually.

    ctypedef long PY_INT64_T  # FIXME: Py2.7+, not defined here but used here

    ctypedef struct PyInterpreterState:
        pass

    ctypedef struct PyThreadState:
        pass

    ctypedef struct PyFrameObject:
        pass

    # This is not actually a struct, but make sure it can never be coerced to
    # an int or used in arithmetic expressions
    ctypedef struct PyGILState_STATE:
        pass

    # The type of the trace function registered using PyEval_SetProfile() and
    # PyEval_SetTrace().
    # Py_tracefunc return -1 when raising an exception, or 0 for success.
    ctypedef int (*Py_tracefunc)(PyObject *, PyFrameObject *, int, PyObject *)

    # The following values are used for 'what' for tracefunc functions
    enum:
        PyTrace_CALL
        PyTrace_EXCEPTION
        PyTrace_LINE
        PyTrace_RETURN
        PyTrace_C_CALL
        PyTrace_C_EXCEPTION
        PyTrace_C_RETURN


    PyInterpreterState * PyInterpreterState_New()
    void PyInterpreterState_Clear(PyInterpreterState *)
    void PyInterpreterState_Delete(PyInterpreterState *)
    PY_INT64_T PyInterpreterState_GetID(PyInterpreterState *)

    PyThreadState * PyThreadState_New(PyInterpreterState *)
    void PyThreadState_Clear(PyThreadState *)
    void PyThreadState_Delete(PyThreadState *)

    PyThreadState * PyThreadState_Get()
    PyThreadState * PyThreadState_Swap(PyThreadState *)  # NOTE: DO NOT USE IN CYTHON CODE !
    PyObject * PyThreadState_GetDict()
    int PyThreadState_SetAsyncExc(long, PyObject *)

    # Ensure that the current thread is ready to call the Python
    # C API, regardless of the current state of Python, or of its
    # thread lock.  This may be called as many times as desired
    # by a thread so long as each call is matched with a call to
    # PyGILState_Release().  In general, other thread-state APIs may
    # be used between _Ensure() and _Release() calls, so long as the
    # thread-state is restored to its previous state before the Release().
    # For example, normal use of the Py_BEGIN_ALLOW_THREADS/
    # Py_END_ALLOW_THREADS macros are acceptable.

    # The return value is an opaque "handle" to the thread state when
    # PyGILState_Ensure() was called, and must be passed to
    # PyGILState_Release() to ensure Python is left in the same state. Even
    # though recursive calls are allowed, these handles can *not* be shared -
    # each unique call to PyGILState_Ensure must save the handle for its
    # call to PyGILState_Release.

    # When the function returns, the current thread will hold the GIL.

    # Failure is a fatal error.
    PyGILState_STATE PyGILState_Ensure()

    # Release any resources previously acquired.  After this call, Python's
    # state will be the same as it was prior to the corresponding
    # PyGILState_Ensure() call (but generally this state will be unknown to
    # the caller, hence the use of the GILState API.)

    # Every call to PyGILState_Ensure must be matched by a call to
    # PyGILState_Release on the same thread.
    void PyGILState_Release(PyGILState_STATE)

    # Routines for advanced debuggers, requested by David Beazley.
    # Don't use unless you know what you are doing!
    PyInterpreterState * PyInterpreterState_Head()
    PyInterpreterState * PyInterpreterState_Next(PyInterpreterState *)
    PyThreadState * PyInterpreterState_ThreadHead(PyInterpreterState *)
    PyThreadState * PyThreadState_Next(PyThreadState *)



from setuptools import setup
from setuptools.extension import Extension
from setuptools.command.build_ext import build_ext
from distutils import log as distutils_logger
from distutils.errors import DistutilsSetupError

import os
import subprocess
import setuptools

ZFP_DOWNLOAD_PATH = 'https://github.com/LLNL/zfp/releases/download/0.5.5/zfp-0.5.5.tar.gz'  # noqa


def download_file(url):
    import requests
    fname = url.split("/")[-1]
    r = requests.get(url)
    with open(fname, 'wb') as f:
        f.write(r.content)


def has_flag(compiler, flagname):
    """Check whether a flag is supported on a compiler."""
    import tempfile
    from distutils.errors import CompileError
    with tempfile.NamedTemporaryFile('w', suffix='.cpp') as f:
        f.write('int main (int argc, char **argv) { return 0; }')
        try:
            compiler.compile([f.name], extra_postargs=[flagname])
        except CompileError:
            return False
    return True


def flag_filter(compiler, *flags):
    """Filter flags, returns list of accepted flags"""
    result = []
    for flag in flags:
        if has_flag(compiler, flag):
            result.append(flag)
    return result


class lazy_cythonize(list):
    def __init__(self, callback):
        self._list, self.callback = None, callback

    def c_list(self):
        if self._list is None:
            self._list = self.callback()
        return self._list

    def __iter__(self):
        for e in self.c_list():
            yield e

    def __getitem__(self, ii):
        return self.c_list()[ii]

    def __len__(self):
        return len(self.c_list())


def extensions():
    import numpy
    from Cython.Build import cythonize
    ext = Extension("pyzfp",
                    sources=["pyzfp.pyx"],
                    include_dirs=['zfp-0.5.5/include',
                                  numpy.get_include()],
                    libraries=["zfp"],  # Unix-like specific,
                    library_dirs=["zfp-0.5.5/lib"],
                    # extra_link_args=['-Wl,-rpath,/usr/local/lib']
                    )
    return cythonize([ext])


class specialized_build_ext(build_ext, object):
    """
    Specialized builder for testlib library
    Code borrowed from: https://stackoverflow.com/a/48641638

    """
    special_extension = "pyzfp"

    def build_extension(self, ext):
        if has_flag(self.compiler, '-fopenmp'):
            for ext in self.extensions:
                ext.extra_compile_args += ['-fopenmp']
                ext.extra_link_args += ['-fopenmp']
            clang = False
        else:
            clang = True

        if ext.name != self.special_extension:
            # Handle unspecial extensions with the parent class' method
            super(specialized_build_ext, self).build_extension(ext)
        else:
            # Handle special extension
            sources = ext.sources
            if sources is None or not isinstance(sources, (list, tuple)):
                raise DistutilsSetupError(
                       "in 'ext_modules' option (extension '%s'), "
                       "'sources' must be present and must be "
                       "a list of source filenames" % ext.name)
            sources = list(sources)
            if len(sources) > 1:
                sources_path = os.path.commonpath(sources)
            else:
                sources_path = os.path.dirname(sources[0])
            sources_path = os.path.realpath(sources_path)
            if not sources_path.endswith(os.path.sep):
                sources_path += os.path.sep

            if not os.path.exists(sources_path) or \
               not os.path.isdir(sources_path):
                raise DistutilsSetupError(
                       "in 'extensions' option (extension '%s'), "
                       "the supplied 'sources' base dir "
                       "must exist" % ext.name)

            download_file(ZFP_DOWNLOAD_PATH)
            command = 'make'
            if clang:
                command += ' OPENMP=0'
            else:
                command += ' OPENMP=1'

            env_vars = ['CC', 'CXX', 'CFLAGS', 'FC']

            for v in env_vars:
                val = os.getenv(v)
                if val is not None:
                    command += ' %s=%s' % (v, val)

            distutils_logger.info('Will execute the following command in ' +
                                  'with subprocess.Popen:' +
                                  '\n{0}'.format(command))
            try:
                output = subprocess.check_output(command,
                                                 cwd=sources_path,
                                                 stderr=subprocess.STDOUT,
                                                 shell=True)
            except subprocess.CalledProcessError as e:
                distutils_logger.info(str(e.output))
                raise

            distutils_logger.info(str(output))

            # After making the library build the c library's python interface
            # with the parent build_extension method
            super(specialized_build_ext, self).build_extension(ext)


with open("README.md", "r") as fh:
    long_description = fh.read()


configuration = {
    'name': 'pyzfp',
    'packages': setuptools.find_packages(),
    'setup_requires': ['cython>=0.17', 'requests', 'numpy', 'setuptools_scm'],
    'ext_modules': lazy_cythonize(extensions),
    'use_scm_version': True,
    'cmdclass': {'build_ext': specialized_build_ext},
    'description': "A python wrapper for the ZFP compression libary",
    'long_description': long_description,
    'long_description_content_type': 'text/markdown',
    'url': 'https://github.com/navjotk/pyzfp',
    'author': "Navjot Kukreja",
    'author_email': 'navjotk@gmail.com',
    'license': 'MIT',
}


setup(**configuration)



import numpy as np
from pyzfp import compress, decompress


def test_compress_decompress():
    a = np.linspace(0, 100, num=1000000).reshape((100, 100, 100))
    tolerance = 0.0000001
    compressed = compress(a, tolerance=tolerance)
    recovered = decompress(compressed, a.shape, a.dtype, tolerance=tolerance)
    assert(a.shape == recovered.shape)
    assert(np.allclose(a, recovered))


def test_dim_order():
    a = np.arange(32, dtype=np.float32).reshape((8, 4))
    compressed = compress(a, rate=8)
    recovered = decompress(compressed[0:16], (4, 4), np.dtype('float32'),
                           rate=8)
    b = np.arange(16, dtype=np.float32).reshape((4, 4))
    assert(np.allclose(recovered, b))


test_compress_decompress()
test_dim_order()
